package de.uni_passau.fim.strategies;
import java.util.LinkedList;
import java.util.List;


/**
 * Representation of answers "(<code>black</code>, <code>white</code>)", 
 * where <code>black</code> is the number of <i>exact</i> hits, i.e., 
 * the number of correct colors in correct positions, 
 * and <code>white</code> is the number of correct hits, i.e., the 
 * number of correct hits in the wrong position.
 * 
 */
public class Answer {
	/** Number of exact hits. */
	private int black;
	
	/** Number of correct hits. */
	private int white;
	
	/** Length of a color code. */
	private int places;
	
	/** The hash code of this object. */
	private int hashCode = 0;
	
	/**
	 * Create an answer (<code>black</code>, <code>white</code>) for a
	 * Mastermind game with <code>length</code> places. 
	 * 
	 * @param black Number of exact hits.
	 * @param white Number of correct hits.
	 * @param length Length of the codewords.
	 */
	public Answer(int black, int white, int length) {
		this.black = black;
		this.white = white;
		this.places = length;
		
		// Compute hashCode.
		int tmpHash = 17;
		tmpHash = 37*tmpHash + black;
		tmpHash = 37*tmpHash + white;
		hashCode = tmpHash;
	}
	
	/** Returns a string representation of the form
	 * <code>(black: x, white: y)</code>. */
	public String toString() {
		return "(black: " + black + ", white: " + white + ")";
	}
	
	/**
	 * Compares this answer with the given object.
	 * @return <code>true</code> only if <code>o</code> is an answer with
	 * the same values for <code>black</code> and <code>withe</code> as 
	 * this answer.
	 */
	public boolean equals(Object o) {
		if (this == o)
			return true;
		if (!(o instanceof Answer))
			return false;
		Answer a = (Answer)o;
		return a.black == this.black && a.white == this.white;
	}
	
	/**
	 * Returns the hash code for this answer.
	 */
	public int hashCode() {
		return hashCode;
	}
	
	/**
	 * Returns a list of all possible answers for the given number of places.
	 */
	public static List<Answer> getAllPossibleAnswers(final int numberOfPlaces) {
		
		// First create all (b,w) combinations with 0 <= b+w <= #places
		List<Answer> answers = new LinkedList<Answer>();
		for (int b=0; b<=numberOfPlaces; b++) {
			for (int w=0; w<=numberOfPlaces-b; w++)
				if (isPossibleAnswer(b, w, numberOfPlaces))
					answers.add(new Answer(b, w, numberOfPlaces));
		}
		
		return answers;
	}
	
	/**
	 * Check if it is possible to have answer (<code>black</code>, 
	 * <code>white</code>) for the given number of places.
	 * 
	 * The only purpose of this method is to demonstrate on a trivial
	 * example the possibility to use methods with a "speaking name" 
	 * instead of long and messy tests (as it would occur in
	 * the above method {@link #getAllPossibleAnswers(int)}) involving 
	 * negation (which for most humans is hard to understand).  
	 */
	private static boolean isPossibleAnswer(int black, int white, 
			int numberOfPlaces) {
		if (white == 1 && black+white == numberOfPlaces)
			return false;
		return true;
	}
	
	/** Return the number of black hits, this answer represents. */
	public int getBlack() {
		return black;
	}
	
	/** Return the number of white hits, this answer represents. */
	public int getWhite() {
		return white;
	}
	
	/** Returns <code>true</true> if this Answer represents a solution, i.e.
	 * all hits are <code>black</code>.  */
	public boolean solved() {
		return black == places;
	}
}
