package de.uni_passau.fim.strategies;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

/**
 * This class provides a collection of utility methods to be used when
 * implementing strategies.
 * 
 * <b>Important note:</b> Some methods take as arguments 
 * <ol>
 * <li> the number of possible colors in the current Mastermind game</li>
 * <li> the number of places, i.e., the length of the codes</li>
 * <li> a boolean that indicates if repetitions in the codes are allowed</li>
 * </ol>
 * 
 * If repetitions are not allowed then it is not possible to have less
 * colors than places, but this is currently not checked by the methods
 * below.
 * 
 */

public class Utils {

	/**
	 * Checks if guess <code>g</code> is consistent with the given configurations.
	 */
	public static boolean isConsistent(Guess g, List<Configuration> configs) {
		if (g == null ||configs == null)
			throw new NullPointerException();
		if (configs.size() == 0)
			return true;
		boolean isConsistent = true;
		for (Configuration configuration : configs) {
			isConsistent &= isConsistent(g, configuration);
		}
		return isConsistent;
	}
	
	/**
	 * Tests if guess <code>g</code> is consistent with a given single configuration.
	 */
	public static boolean isConsistent(Guess g, Configuration config) {
		// System.out.println(config.guess);
		Answer computedAnswer = config.getGuess().computeAnswer(g);
		Answer givenAnswer = config.getAnswer();
		boolean result = givenAnswer.equals(computedAnswer);
		// System.out.println("Guess: " + g + " Rusult: "+ result + " computedAnser = " + computedAnswer + " givenAnswer " + givenAnswer);
		return result;
	}
	
	/**
	 * Filter out all guesses in guessList that are not consistent with configs.
	 * 
	 * @param guessList is modified: Guesses inconsistent with configs are removed.
	 * @param configs List of configurations.
	 * @return The Guesses that were removed from guessList
	 */
	public static List<Guess> filter(List<Guess> guessList, List<Configuration> configs) {
		List<Guess> removedGuesses = new LinkedList<Guess>();
		
		for (Guess guess : guessList) {
			if (!isConsistent(guess, configs)) {
				// System.out.println("Guess " + guess + " is not c with given config.");
				removedGuesses.add(guess);
			}
		}
		
		guessList.removeAll(removedGuesses);
		return removedGuesses;
	}
	
	/**
	 * Filter out all guesses in guessList that are not consistent with a single config.
	 * 
	 * @param guessList is modified: Guesses inconsistent with configs are removed.
	 * @param configs Single configurations.
	 * @return The Guesses that were removed from guessList
	 */
	public static List<Guess> filter(List<Guess> guessList, Configuration config) {
		List<Guess> removedGuesses = new LinkedList<Guess>();
		
		for (Guess guess : guessList) {
			if (!isConsistent(guess, config)) {
				// System.out.println("Guess " + guess + " is not c with given config.");
				removedGuesses.add(guess);
			}
		}
		guessList.removeAll(removedGuesses);
		return removedGuesses;
	}
	
	/**
	 * Creates the list of all guesses that are consistent with a given list
	 * of configurations. 
	 * 
	 * IMPORTANT: For large values of numberOfColors or numberOfPlaces this list can
	 * be HUGE!
	 * 
	 */
	public static List<Guess> createConsistentGuesses(List<Configuration> configs, int numberOfColors, int numberOfPlaces, boolean withRepetitions) {
		List<Guess> l = Utils.createAllGuesses(numberOfColors, numberOfPlaces, withRepetitions);
		filter(l, configs);
		return l;
	}
	
	/**
	 * Creates a list of all guesses.
	 */
	public static List<Guess> createAllGuesses(int numberOfColors, int numberOfPlaces, boolean withRepetitions) {
		List<Guess> guessList = new LinkedList<Guess>();
		if (numberOfPlaces <= 0) 
			throw new RuntimeException("Cannot create a guess with 0 or less places.");
		for (int i=0; i<numberOfPlaces; i++) {
			
			if (i==0) {
				for (int cs=0; cs<numberOfColors; cs++){
					int[] tmp = new int[numberOfPlaces];
					tmp[0] = cs;
					guessList.add(new Guess(tmp, numberOfColors));
				}
			} else {
				List<Guess> newGuessList = new LinkedList<Guess>();
				for (int cs=0; cs<numberOfColors; cs++) {
					List<Guess> tmpGuessList = copyGuessList(guessList);
					List<Guess> removeList = new LinkedList<Guess>();
					for (Guess guess : tmpGuessList) {
						guess.colors[i] = cs;
						if (!withRepetitions) {
							boolean isContained = false;
							for (int counter=0; counter<i;counter++)
								isContained |= (guess.colors[counter] == cs);
							if (isContained)
								removeList.add(guess);
						}
					}
					tmpGuessList.removeAll(removeList);
					newGuessList.addAll(tmpGuessList);
				}
				guessList.clear();
				guessList.addAll(newGuessList);
			}
		}
		return guessList;
	}
	
	/**
	 * Creates a deep copy of <code>src</code> in 
	 * <code>dest</code>. Note, <code>dest</code> will be cleared first.
	 */
	public static void copyGuessList(List<Guess> src, List<Guess> dest) {
		dest.clear();
		for (Guess guess : src) {
			dest.add(guess.clone());
		}
	}
	
	/**
	 * Returns a deep copy of <code>src</code>.
	 * @return A new list which is a copy of <code>src</code>.
	 */
	public static List<Guess> copyGuessList(List<Guess> src) {
		List<Guess> newList = new LinkedList<Guess>();
		copyGuessList(src, newList);
		return newList;
	}
	
	/**
	 * Create a list of length size of random guesses.
	 * @param size The size of the returned list.
	 */
	public static List<Guess> createRandomGuessList(int size, int colors, int places, boolean withRepetition) {
		List<Guess> g = new LinkedList<Guess>();
		for (int i=0; i<size; i++) {
			g.add(createRandomGuess(colors, places, withRepetition));
		}
		return g;
	}
	
	/**
	 * Create a list of length size of random genes.
	 * @param size The size of the returned list.
	 */
	public static List<Gene> createRandomGeneList(int size, int colors, int places, boolean withRepetition) {
		List<Gene> g = new LinkedList<Gene>();
		for (int i=0; i<size; i++) {
			Guess tGuess = createRandomGuess(colors, places, withRepetition);
			g.add(new Gene(tGuess));
		}
		return g;
	}
	
	/**
	 * Create a single random guess.
	 */
	public static Guess createRandomGuess(int colors, int places, boolean withRepetition) {
		if (!Utils.isValidCPRCombination(colors, places, withRepetition))
			throw new RuntimeException("Utils.createRandomGuess(...): Unsupported combination of colors/places/withRepetition!");
		
		// The array containing the random colors.
		int[] c = new int[places];
		// Array where usedColors[i] == 0 only if color i was already
		// used for c when working without repetitions.
		int[] usedColors = new int[colors];
		Arrays.fill(usedColors, 0);
		
		Random rg = new Random();
		for (int i=0; i<c.length; i++) {
			int nextColor = rg.nextInt(colors);
			if (withRepetition) {
				c[i] = nextColor;
			} else {
				if (usedColors[nextColor] == 0) {
					c[i] = nextColor;
					usedColors[nextColor]++;
				} else 
					i--;
			}
		}
		
		return new Guess(c, colors);
	}
	
	public static boolean isValidCPRCombination(int colors, int places, boolean withRepetition) {
		if (withRepetition) {
			return colors > 0 && places > 0;
		} else {
			return colors > 0 && places > 0 && colors >= places;
		}
	}
	
	/**
	 * Compute the factorial of n.
	 */
	public static int factorial(int n) {
		if (n <= 1) 
			return 1;
		else
			return n*factorial(n-1);
	}
	
	/**
	 * Returns the binomial coefficient n over k.
	 */
	public static int binomial(int n, int k) {
		if (n<=0 || k<=0 || k>n || k==n )
			return 1;
		
		int ctr = 2;
		int result = n;
		int newK = Math.max(k, n-k);
		for (int i=n-1; i>newK; i--) {
			result *= i;
			result /= ctr;
			ctr++;
		}
		return result;
	}
}
